import tornado.ioloop
import time
import thread

from monocle import launch


class Task(object):
    def __init__(self, tornado_ioloop, timeout):
        self._timeout = timeout
        self._tornado_ioloop = tornado_ioloop

    def cancel(self):
        self._tornado_ioloop.remove_timeout(self._timeout)


class EventLoop(object):
    def __init__(self):
        self._tornado_ioloop = tornado.ioloop.IOLoop.instance()
        self.READ = self._tornado_ioloop.READ
        self._thread_ident = thread.get_ident()

    def queue_task(self, delay, callable, *args, **kw):
        def task():
            return launch(callable, *args, **kw)

        def queue():
            now = time.time()
            timeout = self._tornado_ioloop.add_timeout(now + delay, task)
            return Task(self._tornado_ioloop, timeout)

        if thread.get_ident() != self._thread_ident:
            self._tornado_ioloop.add_callback(queue)
        else:
            return queue()

    def run(self):
        self._tornado_ioloop.start()

    def halt(self):
        self._tornado_ioloop.stop()

    def _add_handler(self, *a, **k):
        self._tornado_ioloop.add_handler(*a, **k)

    def _remove_handler(self, *a, **k):
        self._tornado_ioloop.remove_handler(*a, **k)

evlp = EventLoop()
queue_task = evlp.queue_task
run = evlp.run
halt = evlp.halt



from __future__ import print_function
import sys
import code
import readline
import atexit
import os
import traceback
import inspect
from threading import Thread

import monocle
from monocle import Return, _o
if __name__ != '__main__':
    from monocle.stack import eventloop

from monocle.callback import Callback


# it's annoying to ever see these warnings at the repl, so tolerate a lot
monocle.core.blocking_warn_threshold = 10000


class HistoryConsole(code.InteractiveConsole):
    def __init__(self, locals=None, filename="<console>",
                 histfile=os.path.expanduser("~/.console-history")):
        code.InteractiveConsole.__init__(self, locals, filename)
        self.init_history(histfile)

    def init_history(self, histfile):
        readline.parse_and_bind("tab: complete")
        if hasattr(readline, "read_history_file"):
            try:
                readline.read_history_file(histfile)
            except IOError:
                pass
            atexit.register(self.save_history, histfile)

    def save_history(self, histfile):
        readline.write_history_file(histfile)


@_o
def repl(debug=True):
    print("Monocle", monocle.VERSION, "/", "Python", sys.version)
    print('Type "help", "copyright", "credits" or "license" for more information.')
    print("You can yield to Monocle oroutines at the prompt.")
    if debug:
        print("Use cont() to continue.")
    ic = HistoryConsole()
    gs = dict(globals())

    done = {'val': False}

    def cont():
        done['val'] = True

    ls = {}
    if debug:
        frame = inspect.currentframe()
        ls = frame.f_back.f_back.f_back.f_back.f_locals
        ls['cont'] = cont

    source = ""
    while not done['val']:
        try:
            if source:
                source += "\n"

            cb = Callback()

            def wait_for_input():
                try:
                    prompt = ">>> "
                    if source:
                        prompt = "... "
                    s = ic.raw_input(prompt)
                except EOFError:
                    eventloop.queue_task(0, eventloop.halt)
                    return
                eventloop.queue_task(0, cb, s)

            Thread(target=wait_for_input).start()
            source += yield cb

            if "\n" in source and not source.endswith("\n"):
                continue

            try:
                _c = code.compile_command(source)
                if not _c:
                    continue
                eval(_c, gs, ls)
            except SyntaxError as e:
                if "'yield' outside function" not in str(e):
                    raise

                # it's a yield!

                try:
                    core_hack_source = "    __r = (" + source.replace("\n", "\n    ") + ")"
                    hack_source = "def __g():\n" + core_hack_source + "\n    yield Return(locals())\n\n"
                    _c = code.compile_command(hack_source)
                except SyntaxError:
                    # maybe the return value assignment wasn't okay
                    core_hack_source = "    " + source.replace("\n", "\n    ")
                    hack_source = "def __g():\n" + core_hack_source + "\n    yield Return(locals())\n\n"
                    _c = code.compile_command(hack_source)

                if not _c:
                    continue

                # make the locals global so __g can reach them
                g_gs = dict(gs)
                g_gs.update(ls)
                eval(_c, g_gs, ls)

                # now monoclize it and get the callback
                _c = code.compile_command("monocle.o(__g)()", symbol="eval")
                cb = eval(_c, gs, ls)
                ls.pop('__g')
                #print "=== waiting for %s ===" % cb
                g_ls = yield cb
                if '__r' in g_ls:
                    r = g_ls.pop('__r')
                    ls['_'] = r
                    if r:
                        print(r)
                ls.update(g_ls)
        except Exception:
            traceback.print_exc()

        source = ""


if __name__ == '__main__':
    monocle.init(sys.argv[1])
    from monocle.stack import eventloop
    monocle.launch(repl, debug=False)
    eventloop.run()



from __future__ import absolute_import
# This code is based heavily on inlineCallbacks from Twisted 10.0, see LICENSE.

import sys
import types
import logging
import traceback
import time
import inspect
import os.path
import functools

from .callback import Callback, defer

try:
    from twisted.python.failure import Failure as TwistedFailure
    from twisted.internet.defer import Deferred as TwistedDeferred
except ImportError:
    class TwistedFailure:
        pass

    class TwistedDeferred:
        pass

log = logging.getLogger("monocle")

blocking_warn_threshold = 500  # ms
tracebacks_elide_internals = True


class Return(object):
    def __init__(self, *args):
        # mimic the semantics of the return statement
        if len(args) == 0:
            self.value = None
        elif len(args) == 1:
            self.value = args[0]
        else:
            self.value = args

    def __repr__(self):
        return "<%s.%s object at 0x%x; value: %s>" % (self.__class__.__module__,
                                                      self.__class__.__name__,
                                                      id(self),
                                                      repr(self.value))


class InvalidYieldException(Exception):
    pass


def is_eventloop_stack(stack):
    this_dir = os.path.dirname(__file__)
    for file, line, context, code in stack:
        if (file.startswith(this_dir) and
            file.endswith("eventloop.py") and
            context == 'run'):
            return True
    return False


def format_stack_lines(stack, elide_internals=tracebacks_elide_internals):
    eliding = False
    lines = []
    for file, line, context, code in stack:
        this_file = __file__
        if this_file.endswith('.pyc'):
            this_file = this_file[:-1]
        if not file == this_file or not elide_internals:
            eliding = False
            lines.append("  File %s, line %s, in %s\n    %s" %
                         (file, line, context, code))
        else:
            if not eliding:
                eliding = True
                lines.append("  -- eliding monocle internals --")
    return lines


def format_tb(e, elide_internals=tracebacks_elide_internals):
    s = ""
    for i, (tb, caller_stack) in enumerate(reversed(e._monocle['tracebacks'])):
        lines = tb.split('\n')

        first = lines[0]  # "Traceback (most recent call last)"

        end_offset = -1
        while (end_offset >= -len(lines) and
               not lines[end_offset].startswith('  File ')):
            end_offset -= 1
        end_offset += 2
        last = lines[end_offset]  # Line(s) describing the exception

        caller_stack_lines = []
        if caller_stack and not is_eventloop_stack(caller_stack):
            caller_stack_lines = format_stack_lines(caller_stack, elide_internals)

        # 3 because of the "Traceback (most recent call last)" line,
        # plus two lines describing the g.throw that got us the
        # exception
        lines = caller_stack_lines + lines[3:end_offset]

        if caller_stack and is_eventloop_stack(caller_stack):
            if i + 1 < len(e._monocle['tracebacks']):
                lines += ["  -- trampolined off eventloop --"]
            else:
                lines += ["  -- trampolined off eventloop, into caller stack --"]
                # the last one is details on how we got called
                if elide_internals:
                    caller_stack = caller_stack[2:]
                lines += format_stack_lines(caller_stack, elide_internals)

        s += "\n" + '\n'.join(lines)
    return first + s + "\n" + last


def _append_traceback(e, tb, caller_stack):
    if not hasattr(e, "_monocle"):
        e._monocle = {'tracebacks': []}
    e._monocle['tracebacks'].append((tb, caller_stack))
    return e


def _add_monocle_tb(e, bounceback=False):
    tb = traceback.format_exc()
    caller_stack = None
    if bounceback:
        caller_stack = traceback.extract_stack()

    # if it's not an eventloop stack, the first one we get is
    # comprehensive and future ones are higher up the stack.  if it is
    # an eventloop stack, we need to add it to reconstruct how we got
    # to the first one.
    if (not caller_stack or
        is_eventloop_stack(caller_stack) or
        not hasattr(e, "_monocle")):
        _append_traceback(e, tb, caller_stack)
    return e


def _add_twisted_tb(f):
    tb = f.getTraceback(elideFrameworkCode=tracebacks_elide_internals)
    return _append_traceback(f.value, tb, None)


def _monocle_chain(to_gen, g, callback):
    # This function is complicated by the need to prevent unbounded recursion
    # arising from repeatedly yielding immediately ready callbacks.  This while
    # loop solves that by manually unfolding the recursion.

    while True:
        try:
            # Send the last result back as the result of the yield expression.
            start = time.time()
            try:
                if isinstance(to_gen, Exception):
                    from_gen = g.throw(type(to_gen), to_gen)
                elif isinstance(to_gen, TwistedFailure):
                    from_gen = g.throw(to_gen.type, to_gen.value, to_gen.tb)
                else:
                    from_gen = g.send(to_gen)
            finally:
                duration = (time.time() - start) * 1000
                if duration > blocking_warn_threshold:
                    if inspect.isframe(g.gi_frame):
                        fi = inspect.getframeinfo(g.gi_frame)
                        log.warn(
                            "oroutine '%s' blocked for %dms before %s:%s", g.__name__, duration, fi.filename, fi.lineno)
                    else:
                        log.warn(
                            "oroutine '%s' blocked for %dms", g.__name__, duration)
        except StopIteration:
            # "return" statement (or fell off the end of the generator)
            from_gen = Return()
        except Exception as e:
            callback(_add_monocle_tb(e, bounceback=(e == to_gen)))
            return callback

        if isinstance(from_gen, Return):
            try:
                g.close()
            except Exception as e:
                callback(_add_monocle_tb(e))
            else:
                callback(from_gen.value)
            return callback
        elif not isinstance(from_gen, Callback):
            if isinstance(from_gen, TwistedDeferred):
                cb = Callback()
                from_gen.addCallbacks(cb, lambda f: cb(_add_twisted_tb(f)))
                from_gen = cb
            else:
                e = InvalidYieldException(
                    "Unexpected value '%s' of type '%s' yielded from o-routine '%s'.  "
                    "O-routines can only yield Callback and Return types." % (from_gen, type(from_gen), g))
                return _monocle_chain(e, g, callback)

        if not hasattr(from_gen, 'result'):
            def gotResult(r):
                _monocle_chain(r, g, callback)
            from_gen.add(gotResult)
            return callback

        to_gen = from_gen.result


def maybeCallbackGenerator(f, *args, **kw):
    try:
        result = f(*args, **kw)
    except Exception as e:
        return defer(_add_monocle_tb(e))

    if isinstance(result, types.GeneratorType):
        return _monocle_chain(None, result, Callback())
    elif isinstance(result, Callback):
        return result
    elif isinstance(result, TwistedDeferred):
        cb = Callback()
        result.addCallbacks(cb, lambda f: cb(_add_twisted_tb(f)))
        return cb
    return defer(result)


# @_o
def _o(f):
    """
    monocle helps you write Callback-using code that looks like a regular
    sequential function.  For example::

        @_o
        def foo():
            result = yield makeSomeRequestResultingInCallback()
            print result

    When you call anything that results in a Callback, you can simply yield it;
    your generator will automatically be resumed when the Callback's result is
    available. The generator will be sent the result of the Callback with the
    'send' method on generators, or if the result was a failure, 'throw'.

    Your coroutine-enabled generator will return a Callback object,
    which will result in the return value of the generator (or will
    fail with a failure object if your generator raises an unhandled
    exception). Note that you can't use "return result" to return a
    value; use "yield Return(result)" instead. Falling off the end of
    the generator, or simply using "return" will cause the Callback to
    have a result of None.  Yielding anything other and a Callback or
    a Return is not allowed, and will raise an exception.

    The Callback returned from your generator will call back with an
    exception if your generator raised an exception::

        @_o
        def foo():
            result = yield makeSomeRequestResultingInCallback()
            if result == 'foo':
                # this will become the result of the Callback
                yield Return('success')
            else:
                # this too
                raise Exception('fail')
    """
    @functools.wraps(f)
    def unwindGenerator(*args, **kwargs):
        return maybeCallbackGenerator(f, *args, **kwargs)
    return unwindGenerator
o = _o


def log_exception(msg="", elide_internals=tracebacks_elide_internals):
    e = sys.exc_info()[1]

    if hasattr(e, '_monocle'):
        log.error("%s\n%s", msg, format_tb(e, elide_internals=elide_internals))
    else:
        log.exception(msg)


@_o
def launch(oroutine, *args, **kwargs):
    try:
        cb = oroutine(*args, **kwargs)
        if not isinstance(cb, (Callback, TwistedDeferred)):
            yield Return(cb)

        r = yield cb
        yield Return(r)
    except GeneratorExit:
        raise
    except Exception:
        log_exception(elide_internals=kwargs.get('elide_internals',
                                                 tracebacks_elide_internals))



try:
    from mock import CallableMixin, Mock, MagicMock
except ImportError:
    raise ImportError('monocle_mock requires mock as an additional dependency')

from monocle.callback import Callback


class _MonocleCallableMixin(CallableMixin):
    def _mock_call(self, *args, **kwargs):
        ''' Mock seems to run all returns through this method, so we intercept
        the return value and wrap it in monocle magic. '''
        result = super(_MonocleCallableMixin, self)._mock_call(*args, **kwargs)
        # Special case for when a person tries to print a MonocleMock directly.
        # In this case the result is a string, and the content is something
        # like <MonocleMock id='4470907728'>. So we look for that particular
        # string and do NOT wrap it in a Monocle callback.
        if (isinstance(result, str) and 
            (result.startswith('<MonocleMock ') or
             result.startswith('<MagicMonocleMock '))):
            return result

        cb = Callback()
        cb(result)
        return cb


class MonocleMock(_MonocleCallableMixin, Mock):
    ''' Use this in the same way you would a regular Python Mock, but for
    monocle'd methods. For example:

    THE OLD WAY:
    @test
    @_o
    def test_something_with_a_yield():
        my_mock = Mock()
        my_mock.chickens.return_value = "cluck"
        result = yield my_mock.chickens()        # <----- THIS FAILS!
        assert "cluck" == result

    THE NEW WAY:
    @test
    @_o
    def test_something_with_a_yield():
        my_mock = MonocleMock()                  # note difference in name
        my_mock.chickens.return_value = "cluck"
        result = yield my_mock.chickens()        # <----- THIS WORKS!
        assert "cluck" == result
    '''
    pass


class MagicMonocleMock(_MonocleCallableMixin, MagicMock):
    ''' Same as above, but also handles "magic" methods (such as __gte__,
    __lte__, etc...)
    '''
    pass



import new

from monocle.stack.eventloop import queue_task
from monocle.callback import Callback


def next_tick():
    cb = Callback()
    cb(None)
    return cb


def immediate(val):
    cb = Callback()
    cb(val)
    return cb


def delayed(seconds, val):
    cb = Callback()
    queue_task(seconds, cb, val)
    return cb


def sleep(seconds):
    cb = Callback()
    queue_task(seconds, cb, None)
    return cb


def monkeypatch(cls):
    def decorator(f):
        orig_method = None
        method = getattr(cls, f.__name__, None)
        if method:
            orig_method = lambda *a, **k: method(*a, **k)

        def g(*a, **k):
            return f(orig_method, *a, **k)

        g.__name__ = f.__name__
        setattr(cls, f.__name__,
                new.instancemethod(g, None, cls))
    return decorator



# -*- coding: utf-8 -*-
#
# by Steven Hazel

import socket
import asyncore

from monocle import _o, Return, launch
from monocle.callback import Callback
from monocle.stack.network import Connection, ConnectionLost
from monocle.asyncore_stack.eventloop import evlp


class _Connection(asyncore.dispatcher_with_send):
    def __init__(self, sock=None, evlp=evlp):
        asyncore.dispatcher_with_send.__init__(self, sock=sock, map=evlp._map)
        self.max_buffer_size = 104857600
        self.buffer = ""
        self.read_cb = None
        self.connect_cb = Callback()

    def attach(self, connection):
        self._write_flushed = connection._write_flushed
        self._closed = connection._closed

    def readable(self):
        return self.read_cb is not None

    def handle_connect(self):
        cb = self.connect_cb
        self.connect_cb = None
        cb(None)

    def handle_read(self):
        self.buffer += self.recv(8192)
        if len(self.buffer) >= self.max_buffer_size:
            # Reached maximum read buffer size
            self.disconnect()
            return
        # it's possible recv called handle_close
        if self.read_cb is not None:
            read_cb = self.read_cb
            self.read_cb = None
            read_cb(None)

    def handle_close(self):
        self.close()
        # XXX: get a real reason from asyncore
        reason = IOError("Connection closed")
        self._closed(reason)

    def initiate_send(self):
        asyncore.dispatcher_with_send.initiate_send(self)
        if len(self.out_buffer) == 0:
            self._write_flushed()

    # functions to support the StackConnection interface

    def write(self, data):
        self.send(data)

    def resume(self):
        self.read_cb = Callback()

    def reading(self):
        return self.readable()

    def closed(self):
        return not self.connected

    def disconnect(self):
        self.handle_close()


class _ListeningConnection(asyncore.dispatcher):
    def __init__(self, handler, evlp=evlp):
        asyncore.dispatcher.__init__(self, map=evlp._map)
        self.handler = handler

    def handle_accept(self):
        (conn, addr) = self.accept()
        connection = Connection(_Connection(sock=conn))
        connection._stack_conn.attach(connection)
        self.handler(connection)


class Service(object):
    def __init__(self, handler, port, bindaddr="", backlog=128, evlp=evlp):
        @_o
        def _handler(s):
            try:
                yield launch(handler, s)
            finally:
                s.close()
        self.handler = _handler
        self.evlp = evlp
        self.port = port
        self.bindaddr = bindaddr
        self.backlog = backlog
        self._conn = _ListeningConnection(self.handler, evlp=evlp)
        self._conn.create_socket(socket.AF_INET, socket.SOCK_STREAM)
        self._conn.set_reuse_addr()

    @_o
    def stop(self):
        self._conn.close()


class Client(Connection):
    @_o
    def connect(self, host, port, evlp=evlp):
        self._stack_conn = _Connection(evlp=evlp)
        self._stack_conn.create_socket(socket.AF_INET, socket.SOCK_STREAM)
        self._stack_conn.connect((host, port))
        self._stack_conn.attach(self)
        yield self._stack_conn.connect_cb


def add_service(service):
    service._conn.bind((service.bindaddr, service.port))
    service._conn.listen(service.port)



# -*- coding: utf-8 -*-
#
# by Steven Hazel

import os
import logging

from monocle import _o, Return, VERSION, log_exception
from monocle.stack.network.http import HttpHeaders, HttpRequest, HttpRouter, extract_response
from monocle.twisted_stack.network import Service, SSLService

from twisted.web import server, resource

log = logging.getLogger("monocle.twisted_stack.network.http")


class _HttpServerResource(resource.Resource):
    isLeaf = 1

    def __init__(self, handler, max_body_str_len):
        self.handler = handler
        self.max_body_str_len = max_body_str_len

    def render(self, request):
        @_o
        def _handler(twisted_request):
            try:
                headers = HttpHeaders()
                for k, vs in twisted_request.requestHeaders.getAllRawHeaders():
                    for v in vs:
                        headers.add(k, v)

                # decode UTF-8 keys, matching tornado arg parsing behavior
                args = dict(
                    [(k.decode('utf-8'), v)
                     for k, v
                     in twisted_request.args.iteritems()])

                body = None
                if hasattr(twisted_request.content, "getvalue"):
                    body = twisted_request.content.getvalue()
                else:
                    twisted_request.content.seek(0, os.SEEK_END)
                    if twisted_request.content.tell() <= self.max_body_str_len:
                        twisted_request.content.seek(0, os.SEEK_SET)
                        body = twisted_request.content.read()
                    twisted_request.content.seek(0, os.SEEK_SET)

                request = HttpRequest(
                    proto=twisted_request.clientproto,
                    host=twisted_request.getRequestHostname(),
                    method=twisted_request.method,
                    uri=twisted_request.uri,
                    args=args,
                    remote_ip=twisted_request.getClientIP(),
                    headers=headers,
                    body=body,
                    body_file=twisted_request.content)
                request._twisted_request = twisted_request

                value = yield self.handler(request)
                code, headers, content = extract_response(value)
            except Exception:
                log_exception()
                code, headers, content = 500, {}, "500 Internal Server Error"
            try:
                if twisted_request._disconnected:
                    return

                twisted_request.setResponseCode(code)
                headers.setdefault('Server', 'monocle/%s' % VERSION)
                grouped_headers = {}
                for name, value in headers.iteritems():
                    if name in grouped_headers:
                        grouped_headers[name].append(value)
                    else:
                        grouped_headers[name] = [value]
                for name, value in grouped_headers.iteritems():
                    if isinstance(name, unicode):
                        name = name.encode('ascii')
                    if isinstance(value, unicode):
                        value = value.encode('ascii')
                    twisted_request.responseHeaders.setRawHeaders(name, value)
                twisted_request.write(content)

                # close connections with a 'close' header
                if headers.get('Connection', '').lower() == 'close':
                    twisted_request.channel.persistent = False

                twisted_request.finish()
            except Exception:
                log_exception()
                raise
        _handler(request)
        return server.NOT_DONE_YET


class HttpServer(Service, HttpRouter):
    def __init__(self, port, handler=None, bindaddr="", backlog=128,
                 max_body_str_len=1024 * 1024):
        HttpRouter.__init__(self)
        self.port = port
        self.handler = handler
        self.bindaddr = bindaddr
        self.backlog = backlog
        self._twisted_listening_port = None
        self.factory = server.Site(_HttpServerResource(self.handle_request,
                                                       max_body_str_len))


class HttpsServer(SSLService, HttpRouter):
    def __init__(self, port, ssl_options, handler=None, bindaddr="", backlog=128,
                 max_body_str_len=1024 * 1024):
        HttpRouter.__init__(self)
        self.port = port
        self.ssl_options = ssl_options
        self.handler = handler
        self.bindaddr = bindaddr
        self.backlog = backlog
        self._twisted_listening_port = None
        self.factory = server.Site(_HttpServerResource(self.handle_request,
                                                       max_body_str_len))



# -*- coding: utf-8 -*-
#
# by Steven Hazel

import errno
import socket

try:
    import ssl  # Python 2.6+
except ImportError:
    ssl = None

from tornado.iostream import IOStream, SSLIOStream

from monocle import _o, launch
from monocle.callback import Callback
from monocle.stack.network import Connection
from monocle.tornado_stack.eventloop import evlp
from monocle.util import monkeypatch


# monkeypatch IOStream to provide a reactive read
@monkeypatch(IOStream)
def __init__(orig_method, self, *a, **k):
    orig_method(self, *a, **k)
    self._read_some = None


@monkeypatch(IOStream)
def _read_from_buffer(orig_method, self, *a, **k):
    if self._read_some is not None:
        if self._read_buffer_size > 0:
            callback = self._read_callback
            self._read_callback = None
            self._read_some = None
            self._run_callback(callback, self._consume(self._read_buffer_size))
            return True
    else:
        return orig_method(self, *a, **k)


@monkeypatch(IOStream)
def read_some(orig_method, self, callback):
    """Call callback when we read some bytes."""
    from tornado import stack_context
    assert not self._read_callback, "Already reading"
    self._read_some = True
    self._read_callback = stack_context.wrap(callback)
    while True:
        if self._read_from_buffer():
            return
        self._check_closed()
        if self._read_to_buffer() == 0:
            break
    self._add_io_state(self.io_loop.READ)


class _Connection(object):

    def __init__(self, iostream):
        self.iostream = iostream
        self.iostream.set_close_callback(self._close_called)
        self.read_cb = None
        self.connect_cb = None

    def attach(self, connection):
        self._write_flushed = connection._write_flushed
        self._closed = connection._closed

    def connect(self, address):
        cb = Callback()
        self.connect_cb = cb
        self.iostream.connect(address, self._connect_complete)
        return cb

    def read_some(self):
        cb = Callback()
        self.read_cb = cb
        self.iostream.read_some(self._read_complete)
        return cb

    def read(self, size):
        cb = Callback()
        self.read_cb = cb
        self.iostream.read_bytes(size, self._read_complete)
        return cb

    def read_until(self, s):
        cb = Callback()
        self.read_cb = cb
        self.iostream.read_until(s, self._read_complete)
        return cb

    def _connect_complete(self, result=None):
        cb = self.connect_cb
        self.connect_cb = None
        if cb:
            cb(result)

    def _read_complete(self, result):
        cb = self.read_cb
        self.read_cb = None
        cb(result)

    def _close_called(self, reason=None):
        # XXX: get a real reason from Tornado
        if reason is None:
            reason = IOError("Connection closed")
        self._closed(reason)

    # functions to support the StackConnection interface

    def write(self, data):
        self.iostream.write(data, self._write_flushed)

    def resume(self):
        self.iostream.resume()

    def reading(self):
        return self.iostream.reading()

    def closed(self):
        return self.iostream.closed()

    def disconnect(self):
        self.iostream.close()


# A Tornado-specific connection which passes through to Tornado's read functions
class TornadoConnection(Connection):

    def read_some(self):
        self._check_reading()
        cb = self._stack_conn.read_some()
        self._queue_timeout(cb)
        return cb

    def read(self, size):
        self._check_reading()
        cb = self._stack_conn.read(size)
        self._queue_timeout(cb)
        return cb

    def read_until(self, s):
        self._check_reading()
        cb = self._stack_conn.read_until(s)
        self._queue_timeout(cb)
        return cb


class Service(object):
    def __init__(self, handler, port, bindaddr="", backlog=128):
        @_o
        def _handler(s):
            try:
                yield launch(handler, s)
            finally:
                s.close()
        self.handler = _handler
        self.port = port
        self.bindaddr = bindaddr
        self.backlog = backlog
        self.ssl_options = None
        self._evlp = None
        self._sock = socket.socket()
        self._sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self._sock.setblocking(0)
        self._sock.bind((self.bindaddr, self.port))
        self._sock.listen(self.backlog)

    def _connection_ready(self, fd, events):
        while True:
            try:
                s, address = self._sock.accept()
            except socket.error as e:
                if e[0] not in (errno.EWOULDBLOCK, errno.EAGAIN):
                    raise
                return
            iostream = self._make_iostream(s)
            if not iostream:
                continue
            connection = TornadoConnection(_Connection(iostream))
            connection._stack_conn.attach(connection)
            self.handler(connection)

    def _add(self, evlp):
        self._evlp = evlp
        self._evlp._add_handler(self._sock.fileno(),
                                self._connection_ready,
                                self._evlp.READ)

    def _make_iostream(self, s):
        return IOStream(s)

    @_o
    def stop(self):
        if self._evlp:
            self._evlp._remove_handler(self._sock.fileno())


class SSLService(Service):

    def __init__(self, handler, port, bindaddr="", backlog=128,
                 ssl_options=None):
        if ssl_options is None:
            ssl_options = {}
        Service.__init__(self, handler, port, bindaddr, backlog)
        self.ssl_options = ssl_options

    def _make_iostream(self, s):
        assert ssl, "Python 2.6+ and OpenSSL required for SSL"
        try:
            s = ssl.wrap_socket(s,
                                server_side=True,
                                do_handshake_on_connect=False,
                                **self.ssl_options)
        except ssl.SSLError as err:
            if err.args[0] == ssl.SSL_ERROR_EOF:
                s.close()
                return
            else:
                raise
        except socket.error as err:
            if err.args[0] == errno.ECONNABORTED:
                s.close()
                return
            else:
                raise
        return SSLIOStream(s)


class Client(TornadoConnection):
    def __init__(self, *args, **kwargs):
        Connection.__init__(self, *args, **kwargs)
        self.ssl_options = None

    @_o
    def connect(self, host, port):
        s = socket.socket()
        iostream = self._make_iostream(s)
        self._stack_conn = _Connection(iostream)
        self._stack_conn.attach(self)
        self._stack_conn.connect((host, port))

        if self.timeout is not None:
            self._queue_timeout(self._stack_conn.connect_cb)

        yield self._stack_conn.connect_cb

    def _make_iostream(self, s):
        return IOStream(s)


class SSLClient(Client):

    def __init__(self, ssl_options=None):
        if ssl_options is None:
            ssl_options = {}
        Connection.__init__(self)
        self.ssl_options = ssl_options

    def _make_iostream(self, s):
        return SSLIOStream(s, ssl_options=self.ssl_options)


def add_service(service, evlp=evlp):
    return service._add(evlp)



import urlparse
import collections
import re
import urllib2
import time
import logging
import base64
import Cookie

from monocle import _o, Return, log_exception, VERSION
from monocle.stack.network import ConnectionLost, Client

try:
    from monocle.stack.network import SSLClient
except Exception:
    pass

log = logging.getLogger(__name__)


class HttpException(Exception):
    pass


# a unique value to let us check when a default was not set to anything else
class _NotSetFlag(object):
    pass


class HttpHeaders(collections.MutableMapping):

    def __init__(self, headers=None):
        self.headers = []
        self.keyset = set()
        if hasattr(headers, 'iteritems'):
            for k, v in headers.iteritems():
                self.add(k, v)
        else:
            for k, v in headers or []:
                self.add(k, v)

    def __len__(self):
        return len(self.headers)

    def keys(self):
        return [k for k, v in self.headers]

    def add(self, key, value):
        key = key.lower()
        self.keyset.add(key)
        self.headers.append((key, value))

    def items(self):
        return self.headers

    def __iter__(self):
        return (k for k, v in self.headers)

    def iteritems(self):
        return (x for x in self.headers)

    def __repr__(self):
        return "<%s %s>" % (self.__class__.__name__, repr(self.headers))

    def get_list(self, key, default=_NotSetFlag):
        key = key.lower()
        if key not in self.keyset:
            if default == _NotSetFlag:
                raise KeyError(key)
            else:
                return default
        vals = [v for k, v in self.headers if k == key]
        return vals

    def __getitem__(self, key):
        vals = self.get_list(key)
        if len(vals) == 1:
            return vals[0]
        else:
            return vals

    def __setitem__(self, key, value):
        try:
            del self[key]
        except KeyError:
            pass
        return self.add(key, value)

    def __delitem__(self, key):
        key = key.lower()
        if key not in self.keyset:
            raise KeyError(key)
        self.keyset.remove(key)
        self.headers = [(k, v) for k, v in self.headers if k != key]


class HttpRequest(object):
    def __init__(self, proto='HTTP/1.0', host=None, method=None,
                 uri=None, args=None, remote_ip=None, headers=None,
                 body=None, body_file=None):
        self.proto = proto
        self.host = host
        self.method = method
        self.uri = uri
        self.remote_ip = remote_ip
        self.headers = headers
        self.body = body
        self.body_file = body_file

        self.path, _, self.query = uri.partition('?')
        self.query_args = urlparse.parse_qs(self.query, keep_blank_values=True)
        self.args = args

        self.cookies = {}
        for cookie in self.headers.get_list("cookie", []):
            try:
                for name, morsel in Cookie.BaseCookie(cookie).iteritems():
                    self.cookies[name] = morsel.value
            except Cookie.CookieError:
                pass

    def __repr__(self):
        return "<%s (%s %s %s, headers=%s)>" % (
            self.__class__.__name__, self.method, self.path, self.proto, self.headers)

    def get_basic_auth(self):
        if 'authorization' not in self.headers:
            return None, None
        auth = self.headers["authorization"]
        try:
            method, b64 = auth.split(" ")
            if method.lower() != "basic":
                return None, None
            username, password = base64.decodestring(b64).split(':', 1)
        except Exception:
            # parsing error; no valid auth
            return None, None
        return username, password


class HttpResponse(object):
    def __init__(self, code, msg=None, headers=None, body=None, proto=None):
        self.code = code
        self.msg = msg
        self.headers = headers or HttpHeaders()
        self.body = body
        self.proto = proto or 'HTTP/1.1'


def parse_headers(lines):
    headers = HttpHeaders()
    for line in lines:
        k, v = line.split(":", 1)
        headers.add(k, v.lstrip())
    return headers


def parse_request(data):
    data = data[:-4]
    lines = data.split("\r\n")
    method, path, proto = lines[0].split(" ", 2)
    headers = parse_headers(lines[1:])
    return method, path, proto, headers


def parse_response(data):
    data = data[:-4]
    lines = data.split("\r\n")
    parts = lines[0].split(" ")
    proto = parts[0]
    code = parts[1]
    if len(parts) > 2:
        msg = parts[2]
    else:
        msg = ""
    headers = parse_headers(lines[1:])
    return proto, code, msg, headers


@_o
def read_request(conn):
    data = yield conn.read_until("\r\n\r\n")
    method, path, proto, headers = parse_request(data)
    body = None
    if method in ["POST", "PUT"] and "Content-Length" in headers:
        cl = int(headers["Content-Length"])
        body = yield conn.read(cl)
    yield Return(method, path, proto, headers, body)


@_o
def write_request(conn, method, path, headers, body=None):
    yield conn.write("%s %s HTTP/1.1\r\n" % (method, path))
    for k, v in headers.iteritems():
        yield conn.write("%s: %s\r\n" % (k, v))
    yield conn.write("\r\n")
    if body:
        yield conn.write(body)


@_o
def read_response(conn):
    data = yield conn.read_until("\r\n\r\n")
    proto, code, msg, headers = parse_response(data)

    proto = proto.lower()
    content_length = int(headers.get('Content-Length', 0))
    body = ""

    # From rfc2616 section 4.4:
    # Messages MUST NOT include both a Content-Length header field and
    # a non-identity transfer-coding. If the message does include a
    # non- identity transfer-coding, the Content-Length MUST be
    # ignored.
    if headers.get('Transfer-Encoding', '').lower() == 'chunked':
        while True:
            line = yield conn.read_until("\r\n")
            line = line[:-2]
            parts = line.split(';')
            chunk_len = int(parts[0], 16)
            body += yield conn.read(chunk_len)
            yield conn.read_until("\r\n")
            if not chunk_len:
                break
    elif content_length:
        body = yield conn.read(content_length)
    elif ((proto == 'http/1.0' and
           not headers.get('Connection', '').lower() == 'keep-alive') or
          (proto == 'http/1.1' and
           headers.get('Connection', '').lower() == 'close')):
        while True:
            try:
                body += yield conn.read_some()
            except ConnectionLost:
                break

    yield Return(HttpResponse(code, msg, headers, body, proto))


@_o
def write_response(conn, resp):
    yield conn.write("%s %s %s\r\n" % (resp.proto.upper(), resp.code, resp.msg))
    for k, v in resp.headers.iteritems():
        yield conn.write("%s: %s\r\n" % (k, v))
    yield conn.write('\r\n')
    if resp.body:
        yield conn.write(resp.body)


class HttpClient(object):
    DEFAULT_PORTS = {'http': 80,
                     'https': 443}

    def __init__(self):
        self.client = None
        self.scheme = None
        self.host = None
        self.port = None
        self._timeout = None

    @property
    def timeout(self):
        return self._timeout

    @timeout.setter
    def timeout(self, value):
        self._timeout = value
        if self.client:
            self.client.timeout = value

    @_o
    def connect(self, host, port,
                scheme='http', timeout=None, is_proxy=False):
        if timeout is not None:
            # this parameter is deprecated
            self.timeout = None

        if self.client and not self.client.is_closed():
            self.client.close()

        if scheme == 'http':
            self.client = Client()
        elif SSLClient and scheme == 'https':
            self.client = SSLClient()
        else:
            raise HttpException('unsupported url scheme %s' % scheme)
        self.scheme = scheme
        self.host = host
        self.port = port
        self.is_proxy = is_proxy
        self.client.timeout = self._timeout
        yield self.client.connect(self.host, self.port)

    @_o
    def request(self, url, headers=None, method='GET', body=None):
        parts = urlparse.urlsplit(url)

        if self.is_proxy:
            host = parts.netloc
            path = url
        else:
            if parts.scheme and parts.scheme != self.scheme:
                raise HttpException("URL is %s but connection is %s" %
                                    (parts.scheme, self.scheme))

            host = parts.netloc
            if not host:
                host = self.host
                if self.port != self.DEFAULT_PORTS[self.scheme]:
                    host += ":%s" % self.port

            path = parts.path
            if parts.query:
                path += '?' + parts.query

        if not headers:
            headers = HttpHeaders()
        headers.setdefault('User-Agent', 'monocle/%s' % VERSION)
        headers.setdefault('Host', host)
        if body is not None:
            headers['Content-Length'] = str(len(body))

        yield write_request(self.client, method, path, headers, body)
        response = yield read_response(self.client)
        yield Return(response)

    def close(self):
        self.client.close()

    def is_closed(self):
        return self.client is None or self.client.is_closed()

    @classmethod
    @_o
    def query(cls, url, headers=None, method='GET', body=None):
        self = cls()
        parts = urlparse.urlsplit(url)
        host = parts.hostname
        port = parts.port or self.DEFAULT_PORTS[parts.scheme]

        if not self.client or self.client.is_closed():
            yield self.connect(host, port, scheme=parts.scheme)
        elif not (self.host, self.port) == (host, port):
            self.client.close()
            yield self.connect(host, port, scheme=parts.scheme)

        try:
            result = yield self.request(url, headers, method, body)
        finally:
            self.close()
        yield Return(result)


# Takes a response return value like:
# "this is a body"
# 404
# (200, "this is a body")
# (200, {"headers": "here"}, "this is a body")
#
# ...and converts that to a full (code, headers, body) tuple.
def extract_response(value):
    if isinstance(value, basestring):
        return (200, HttpHeaders(), value)
    if isinstance(value, int):
        return (value, HttpHeaders(), "")
    if len(value) == 2:
        return (value[0], HttpHeaders(), value[1])
    return value


class HttpRouter(object):
    named_param_re = re.compile(r':([^\/\?\*\:]+)')

    def __init__(self):
        self.routes = collections.defaultdict(list)

    @classmethod
    def path_matches(cls, path, pattern):
        m = pattern.match(path)
        if not m:
            return False, None
        if not m.end() == len(path):
            # must match the whole string
            return False, None
        return True, m.groupdict()

    def mk_decorator(self, method, pattern, add_head=False):
        if not hasattr(pattern, 'match'):
            pattern = re.escape(pattern)
            pattern = pattern.replace(r'\?', '?')
            pattern = pattern.replace(r'\:', ':')
            pattern = pattern.replace(r'\_', '_')
            pattern = pattern.replace(r'\/', '/')
            pattern = pattern.replace(r'\*', '.*')
            pattern = self.named_param_re.sub(r'(?P<\1>[^/]+)', pattern)
            pattern = re.compile("^" + pattern + "$")

        def decorator(f):
            handler = _o(f)
            self.routes[method].append((pattern, handler))
            if add_head:
                self.routes['HEAD'].append((pattern, handler))
            return handler
        return decorator

    def get(self, pattern, add_head=True):
        return self.mk_decorator('GET', pattern, add_head=add_head)

    def post(self, pattern):
        return self.mk_decorator('POST', pattern)

    def put(self, pattern):
        return self.mk_decorator('PUT', pattern)

    def delete(self, pattern):
        return self.mk_decorator('DELETE', pattern)

    def head(self, pattern):
        return self.mk_decorator('HEAD', pattern)

    def options(self, pattern):
        return self.mk_decorator('OPTIONS', pattern)

    def patch(self, pattern):
        return self.mk_decorator('PATCH', pattern)

    def route_match(self, req):
        for pattern, handler in self.routes[req.method]:
            match, kwargs = self.path_matches(urllib2.unquote(req.path),
                                              pattern)
            if match:
                return handler, kwargs
        return None, None

    @_o
    def request_handler_wrapper(self, req, handler, **kwargs):
        resp = yield handler(req, **kwargs)
        yield Return(resp)

    @_o
    def handle_request(self, req):
        before = time.time()
        resp = None

        handler, kwargs = self.route_match(req)
        try:
            if handler:
                resp = yield self.request_handler_wrapper(req, handler, **kwargs)
            elif self.handler:
                resp = yield self.request_handler_wrapper(req, self.handler)
            else:
                resp = (404, {}, "")
        except Exception:
            log_exception()
            resp = (500, {}, "500 Internal Server Error")
        after = time.time()

        content_length = 0
        if len(resp) > 2:
            content_length = len(resp[2])

        log.info("[%s] %s %s %s -> %s (%s bytes, %.0fms); %s",
                 req.remote_ip,
                 req.method, req.path, req.proto,
                 resp[0], content_length, (after - before) * 1000,
                 req.headers.get('user-agent'))

        yield Return(resp)


import monocle

if monocle._stack_name == 'twisted':
    from monocle.twisted_stack.network.http import *
elif monocle._stack_name == 'tornado':
    from monocle.tornado_stack.network.http import *



import sys
import thread
import signal

from monocle import launch

# prefer fast reactors
# FIXME: this should optionally refuse to use slow ones
if 'twisted.internet.reactor' not in sys.modules:
    try:
        from twisted.internet import epollreactor
        epollreactor.install()
    except Exception:
        try:
            from twisted.internet import kqreactor
            kqreactor.install()
        except Exception:
            try:
                from twisted.internet import pollreactor
                pollreactor.install()
            except Exception:
                pass

from twisted.internet import reactor
try:
    from twisted.internet.error import ReactorNotRunning
except ImportError:
    ReactorNotRunning = RuntimeError


# thanks to Peter Norvig
def singleton(object, message="singleton class already instantiated",
              instantiated=[]):
    """
    Raise an exception if an object of this class has been instantiated before.
    """
    assert object.__class__ not in instantiated, message
    instantiated.append(object.__class__)


class Task(object):
    def __init__(self, df):
        self._df = df

    def cancel(self):
        self._df.cancel()


class EventLoop(object):
    def __init__(self):
        singleton(self, "Twisted can only have one EventLoop (reactor)")
        self._halted = False
        self._thread_ident = thread.get_ident()
        self._sig_handlers = {}

        # queue this up as early as possible, so that signal handlers get
        # reinstalled just after the reactor starts up
        self.queue_task(0, self._reinstall_signal_handlers)

    def _preserve_sig_handlers(self):
        """ Obnoxiously, twisted overwrites any event handlers with its own when
        its reactor starts up. Nobody wants that. To make twisted work more like
        other stacks, we preserve any non-default signal handlers so that they
        can be reinstalled after the reactor has started. """
        default_handlers = {None, signal.SIG_DFL, signal.SIG_IGN,
                            signal.default_int_handler}
        i = 1
        while True:
            try:
                handler = signal.getsignal(i)
                if handler not in default_handlers:
                    self._sig_handlers[i] = handler
            except ValueError:
                break
            i += 1

    def _reinstall_signal_handlers(self):
        """ Reinstall preserved signal handlers """
        for sig, handler in self._sig_handlers.iteritems():
            signal.signal(sig, handler)

    def queue_task(self, delay, callable, *args, **kw):
        if thread.get_ident() != self._thread_ident:
            reactor.callFromThread(reactor.callLater,
                                   delay, launch, callable, *args, **kw)
        else:
            df = reactor.callLater(delay, launch, callable, *args, **kw)
            return Task(df)

    def run(self):
        if not self._halted:
            self._thread_ident = thread.get_ident()

            # preserve signal handlers just before reactor startup; a task has
            # already been queued up to reinstall them as quickly as possible
            self._preserve_sig_handlers()

            reactor.run()

    def halt(self):
        try:
            reactor.stop()
        except ReactorNotRunning:
            self._halted = True
            pass

evlp = EventLoop()
queue_task = evlp.queue_task
run = evlp.run
halt = evlp.halt



# -*- coding: utf-8 -*-
#
# by Steven Hazel

import io
import tornado.httpclient
import tornado.httpserver

from httplib import responses
from monocle import _o, Return, VERSION, launch
from monocle.callback import Callback
from monocle.stack.network.http import HttpHeaders, HttpRequest, HttpClient, HttpRouter, extract_response
from monocle.tornado_stack.network import Service


class HttpClient(HttpClient):
    @classmethod
    @_o
    def query(self, url, headers=None, method='GET', body=None):
        _http_client = tornado.httpclient.AsyncHTTPClient()
        req = tornado.httpclient.HTTPRequest(url,
                                             method=method,
                                             headers=headers or {},
                                             body=body,
                                             # XXX: TODO
                                             #request_timeout=self.timeout
                                             )
        cb = Callback()
        _http_client.fetch(req, cb)
        response = yield cb
        yield Return(response)


class HttpServer(Service, HttpRouter):
    def __init__(self, port, handler=None):
        HttpRouter.__init__(self)
        self.handler = handler
        self.port = port

    def _add(self, el):
        @_o
        def _handler(tornado_request):
            try:
                headers = HttpHeaders()
                for k, v in tornado_request.headers.get_all():
                    headers.add(k, v)

                body_file = io.BytesIO(tornado_request.body)
                request = HttpRequest(proto=tornado_request.version,
                                      host=tornado_request.host,
                                      method=tornado_request.method,
                                      uri=tornado_request.uri,
                                      args=tornado_request.arguments,
                                      remote_ip=tornado_request.remote_ip,
                                      headers=headers,
                                      body=tornado_request.body,
                                      body_file=body_file)
                request._tornado_request = tornado_request

                value = yield launch(self.handle_request, request)
                code, headers, content = extract_response(value)
            except Exception:
                code, headers, content = 500, {}, "500 Internal Server Error"
            tornado_request.write("HTTP/1.1 %s %s\r\n" %
                                  (code, responses.get(code, 'Unknown')))
            headers.setdefault('Server', 'monocle/%s' % VERSION)
            headers.setdefault('Content-Length', str(len(content)))
            for name, value in headers.iteritems():
                if isinstance(name, unicode):
                    name = name.encode('ascii')
                if isinstance(value, unicode):
                    value = value.encode('ascii')
                tornado_request.write("%s: %s\r\n" % (name, value))
            tornado_request.write("\r\n")
            tornado_request.write(content)
            tornado_request.finish()
        self._http_server = tornado.httpserver.HTTPServer(
            _handler,
            io_loop=el._tornado_ioloop)
        self._http_server.listen(self.port)



class Callback(object):
    """
    Sort of like Twisted's Deferred, but simplified.  We don't do
    callback chaining, since oroutines replace that mechanism.
    """

    def __init__(self):
        self._handlers = []

    def add(self, handler):
        if hasattr(self, 'result'):
            handler(self.result)
        else:
            if not callable(handler):
                raise TypeError("'%s' object is not callable" % type(handler).__name__)
            self._handlers.append(handler)

    def __call__(self, result):
        assert not hasattr(self, 'result'), "Already called back"
        for handler in self._handlers:
            handler(result)
        self.result = result


def defer(result):
    cb = Callback()
    cb(result)
    return cb



# -*- coding: utf-8 -*-
#
# by Steven Hazel

from monocle.twisted_stack.eventloop import reactor
from twisted.internet.protocol import Factory, Protocol, ClientFactory
from twisted.internet import ssl
from twisted.internet.error import TimeoutError

from monocle import _o, launch
from monocle.callback import Callback
from monocle.stack.network import Connection
from monocle.util import monkeypatch
import twisted
from twisted.protocols.tls import TLSMemoryBIOProtocol

twisted_major_version = int(twisted.__version__.split('.')[0])


# monkeypatch TLSMemoryBIOProtocol to resume reading when necessary for writes
@monkeypatch(TLSMemoryBIOProtocol)
def __init__(orig_method, self, *a, **k):
    orig_method(self, *a, **k)
    self._was_paused = False


def _is_blocked(o):
    if twisted_major_version > 16 or not hasattr(o, '_writeBlockedOnRead'):
        return len(o._appSendBuffer) > 0
    else:
        return o._writeBlockedOnRead


@monkeypatch(TLSMemoryBIOProtocol)
def _write(orig_method, self, bytes):
    was_blocked = _is_blocked(self)
    orig_method(self, bytes)
    if not was_blocked and _is_blocked(self):
        if (self.transport.producer is not None and
            self.transport.producer._producerPaused):
            self._was_paused = True
            self.transport.resumeProducing()


@monkeypatch(TLSMemoryBIOProtocol)
def dataReceived(orig_method, self, bytes):
    was_blocked = _is_blocked(self)
    orig_method(self, bytes)
    if was_blocked and not _is_blocked(self):
        if self._was_paused:
            self.transport.pauseProducing()
            self._was_paused = False


class _Connection(Protocol):
    def __init__(self):
        self.read_cb = None
        self.connect_cb = None

    def attach(self, connection):
        self._write_flushed = connection._write_flushed
        self._closed = connection._closed

    # functions to support Protocol

    def connectionMade(self):
        self.max_buffer_size = 104857600
        self.buffer = ""
        self.transport.pauseProducing()
        if hasattr(self.factory, "handler"):
            connection = Connection(self)
            self.attach(connection)
        self.transport.registerProducer(self, False)
        if hasattr(self.factory, "handler"):
            self.factory.handler(connection)
        if self.connect_cb:
            cb = self.connect_cb
            self.connect_cb = None
            cb(None)

    def dataReceived(self, data):
        self.transport.pauseProducing()
        self.buffer += data
        if len(self.buffer) >= self.max_buffer_size:
            # Reached maximum read buffer size
            self.disconnect()
            return
        read_cb = self.read_cb
        self.read_cb = None
        read_cb(None)

    def connectionLost(self, reason):
        self._closed(reason.value)

    # functions to support IPullProducer

    def resumeProducing(self):
        if self._write_flushed:
            self._write_flushed()

    def stopProducing(self):
        # we just wait for the connection lost event
        pass

    # functions to support the StackConnection interface

    def write(self, data):
        self.transport.write(data)

    def resume(self):
        self.read_cb = Callback()
        self.transport.resumeProducing()

    def reading(self):
        return self.read_cb is not None

    def closed(self):
        return not (self.transport and
                    self.transport.connected and
                    not self.transport.disconnecting)

    def disconnect(self):
        if self.transport:
            self.transport.unregisterProducer()
            self.transport.loseConnection()


class Service(object):
    def __init__(self, handler, port, bindaddr="", backlog=128):
        self.factory = Factory()
        self.factory.protocol = _Connection

        @_o
        def _handler(s):
            try:
                yield launch(handler, s)
            finally:
                s.close()
        self.factory.handler = _handler
        self.port = port
        self.bindaddr = bindaddr
        self.backlog = backlog
        self._twisted_listening_port = None

    def _add(self):
        self._twisted_listening_port = reactor.listenTCP(
            self.port, self.factory,
            backlog=self.backlog,
            interface=self.bindaddr)

    @_o
    def stop(self):
        df = self._twisted_listening_port.stopListening()
        if df:
            yield df


class SSLService(Service):

    def __init__(self, handler, ssl_options, port, bindaddr="", backlog=128):
        Service.__init__(self, handler, port, bindaddr, backlog)
        self.ssl_options = ssl_options

    def _add(self):
        cf = ssl.DefaultOpenSSLContextFactory(self.ssl_options['keyfile'],
                                              self.ssl_options['certfile'])
        self._twisted_listening_port = reactor.listenSSL(
            self.port, self.factory, cf,
            backlog=self.backlog,
            interface=self.bindaddr)


class SSLContextFactory(ssl.ClientContextFactory):

    def __init__(self, ssl_options):
        self.ssl_options = ssl_options

    def getContext(self):
        ctx = ssl.ClientContextFactory.getContext(self)
        if 'certfile' in self.ssl_options:
            ctx.use_certificate_file(self.ssl_options['certfile'])
        if 'keyfile' in self.ssl_options:
            ctx.use_privatekey_file(self.ssl_options['keyfile'])
        return ctx


class Client(Connection):
    def __init__(self, *args, **kwargs):
        Connection.__init__(self, *args, **kwargs)
        self._connection_timeout = "unknown"

    def clientConnectionFailed(self, connector, reason):
        if reason.type == TimeoutError:
            msg = ("connection timed out after %s seconds" %
                   self._connection_timeout)
        else:
            msg = reason.value
        self._closed(msg)

    @_o
    def connect(self, host, port):
        self._connection_timeout = self.timeout
        self._stack_conn = _Connection()
        self._stack_conn.attach(self)
        self._stack_conn.connect_cb = Callback()
        factory = ClientFactory()
        factory.protocol = lambda: self._stack_conn
        factory.clientConnectionFailed = self.clientConnectionFailed
        self._connect_to_reactor(host, port, factory, self._connection_timeout)
        yield self._stack_conn.connect_cb

    def _connect_to_reactor(self, host, port, factory, timeout):
        reactor.connectTCP(host, port, factory, timeout=timeout)


class SSLClient(Client):

    def __init__(self, ssl_options=None):
        if ssl_options is None:
            ssl_options = {}
        Connection.__init__(self)
        self.ssl_options = ssl_options

    def _connect_to_reactor(self, host, port, factory, timeout):
        reactor.connectSSL(host, port, factory,
                           SSLContextFactory(self.ssl_options),
                           timeout=timeout)


def add_service(service):
    return service._add()



import monocle
from monocle import _o, Return
from monocle.callback import Callback, defer
from monocle.stack import eventloop as evlp


class ConnectionLost(Exception):
    pass


class Connection(object):
    """
    Connection takes a stack_conn, which should have this interface:

    def write(data):
        writes the data

    read_cb: callback which is called when a read completes
    connect_cb: callback which is called when the connection completes

    def resume():
        resumes reading

    def reading():
        returns a boolean indicating the current reading state

    def closed():
        returns a boolean indicating the current closed state

    def disconnect():
        closes the connection

    if the read/read_until implementations here are used:

    buffer: string buffer which grows before read_cb is called
    """
    def __init__(self, stack_conn=None):
        self._stack_conn = stack_conn
        self.writing = False
        self.flush_cb = Callback()
        self.write_encoding = 'utf-8'
        self.timeout = None
        self._current_timeout = None

    def _time_out(self, instigating_cb, timeout):
        self._current_timeout = None
        if hasattr(instigating_cb, 'result'):
            return

        current_cbs = [self._stack_conn.read_cb,
                       self.flush_cb,
                       self._stack_conn.connect_cb]
        self._stack_conn.read_cb = None
        self.flush_cb = None
        self._stack_conn.connect_cb = None
        self._stack_conn.disconnect()
        for cb in current_cbs:
            if cb and not hasattr(cb, 'result'):
                cb(ConnectionLost("connection timed out after %s seconds" %
                                  timeout))

    def _queue_timeout(self, cb):
        if self.timeout is not None:
            if self._current_timeout:
                self._current_timeout.cancel()
            self._current_timeout = evlp.queue_task(self.timeout,
                                                    self._time_out,
                                                    cb,
                                                    self.timeout)

    @_o
    def read_some(self):
        self._check_reading()

        if not self._stack_conn.buffer:
            self._check_closed()
            self._stack_conn.resume()
            self._queue_timeout(self._stack_conn.read_cb)
            yield self._stack_conn.read_cb
        tmp = self._stack_conn.buffer
        self._stack_conn.buffer = ""
        yield Return(tmp)

    @_o
    def read(self, size):
        self._check_reading()

        while len(self._stack_conn.buffer) < size:
            self._check_closed()
            self._stack_conn.resume()
            self._queue_timeout(self._stack_conn.read_cb)
            yield self._stack_conn.read_cb
        tmp = self._stack_conn.buffer[:size]
        self._stack_conn.buffer = self._stack_conn.buffer[size:]
        yield Return(tmp)

    @_o
    def read_until(self, s):
        self._check_reading()

        while True:
            size = self._stack_conn.buffer.find(s)
            if size != -1:
                size += len(s)
                break
            self._check_closed()
            self._stack_conn.resume()
            self._queue_timeout(self._stack_conn.read_cb)
            yield self._stack_conn.read_cb
        tmp = self._stack_conn.buffer[:size]
        self._stack_conn.buffer = self._stack_conn.buffer[size:]
        yield Return(tmp)

    def readline(self):
        return self.read_until("\n")

    def write(self, data):
        if isinstance(data, unicode):
            data = data.encode(self.write_encoding)
        self._check_closed()
        if data == "":
            # twisted won't ever tell us we flushed writing ""
            return defer(None)
        self.writing = True
        self._stack_conn.write(data)
        cb = self.flush()
        self._queue_timeout(cb)
        return cb

    def _write_flushed(self, result=None):
        self.writing = False
        cb = self.flush_cb
        self.flush_cb = Callback()
        if cb:
            cb(result)

    def flush(self):
        self._check_closed()
        cb = self.flush_cb
        if not self.writing:
            self._write_flushed()
        return cb

    def _closed(self, reason):
        if self._current_timeout:
            self._current_timeout.cancel()
            self._current_timeout = None
        cl = ConnectionLost(str(reason))
        cl.original = reason
        if self._stack_conn.connect_cb:
            cb = self._stack_conn.connect_cb
            self._stack_conn.connect_cb = None
            cb(cl)
        self._write_flushed(cl)
        if self._stack_conn.read_cb:
            cb = self._stack_conn.read_cb
            self._stack_conn.read_cb = None
            cb(cl)

    def close(self):
        self._stack_conn.disconnect()

    def _check_reading(self):
        if self._stack_conn.reading():
            raise IOError("Already reading")

    def _check_closed(self):
        if self._stack_conn.closed():
            raise IOError("Stream is closed")

    def is_closed(self):
        return self._stack_conn.closed()


if monocle._stack_name == 'twisted':
    from monocle.twisted_stack.network import *
elif monocle._stack_name == 'tornado':
    from monocle.tornado_stack.network import *
elif monocle._stack_name == 'asyncore':
    from monocle.asyncore_stack.network import *



import monocle
import os
if monocle._stack_name == 'twisted' or os.getenv('MONOCLE_STACK') == 'twisted':
    from monocle.twisted_stack.eventloop import *
elif monocle._stack_name == 'tornado' or os.getenv('MONOCLE_STACK') == 'tornado':
    from monocle.tornado_stack.eventloop import *
elif monocle._stack_name == 'asyncore' or os.getenv('MONOCLE_STACK') == 'asyncore':
    from monocle.asyncore_stack.eventloop import *
else:
    raise ImportError(
        "Could not import stack.\n"
        "Ensure you have called monocle.init()\n"
        "or defined MONOCLE_STACK=<stack> in the environment")



from multiprocessing import Pipe

from monocle.stack.multiprocess import launch_proc

log = logging.getLogger(__name__)


class PipeForTwisted(object):
    def __init__(self, pipe):
        self.pipe = pipe
        self.callback = Callback()

    def doRead(self):
        self.callback((False, None))

    def fileno(self):
        return self.pipe.fileno()

    def connectionLost(self, reason):
        self.callback((True, reason))

    def logPrefix(self):
        return "Pipe"

    def doWrite(self):
        self.callback((False, None))


class PipeChannel(object):
    def __init__(self, pipe):
        self.pipe = pipe

    @_o
    def send(self, value):
        w = PipeForTwisted(self.pipe)
        eventloop.reactor.addWriter(w)
        lost, reason = yield w.callback
        eventloop.reactor.removeWriter(w)

        if lost:
            raise Exception("connection lost: %s" % reason)
        self.pipe.send(value)

    @_o
    def recv(self):
        r = PipeForTwisted(self.pipe)
        eventloop.reactor.addReader(r)
        lost, reason = yield r.callback
        eventloop.reactor.removeReader(r)

        if lost:
            raise Exception("connection lost: %s" % reason)
        yield Return(self.pipe.recv())


def launch_proc_with_pipes(target, *args, **kwargs):
    child, parent = Pipe()
    pc = PipeChannel(parent)
    p = launch_proc(target, pc, *args, **kwargs)
    cc = PipeChannel(child)
    return p, cc



import asyncore
import time
import functools

from monocle import launch


class EventLoop(object):
    def __init__(self):
        self._running = True
        self._queue = []
        self._map = {}

    def queue_task(self, delay, callable, *args, **kw):
        now = time.time()
        when = now + delay
        self._queue.append((when, callable, args, kw))
        self._queue.sort(reverse=True)

    def run(self):
        while self._running:
            timeout = 0
            if self._queue:
                next = self._queue[-1][0] - time.time()
                if next <= 0:
                    task = self._queue.pop()
                    launch(task[1], *task[2], **task[3])
                else:
                    timeout = next
            if self._map:
                asyncore.loop(timeout=timeout, use_poll=True, count=1,
                              map=self._map)
            else:
                time.sleep(0.1)

    def halt(self):
        self._running = False

evlp = EventLoop()
queue_task = evlp.queue_task
run = evlp.run
halt = evlp.halt
